# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for all logging methods/classes that don't need the ORM."""
import collections
import contextlib
import logging
import types

__all__ = ('AIIDA_LOGGER', 'override_log_level')

# Custom logging level, intended specifically for informative log messages reported during WorkChains.
# We want the level between INFO(20) and WARNING(30) such that it will be logged for the default loglevel, however
# the value 25 is already reserved for SUBWARNING by the multiprocessing module.
LOG_LEVEL_REPORT = 23

# Add the custom log level to the :mod:`logging` module and add a corresponding report logging method.
logging.addLevelName(LOG_LEVEL_REPORT, 'REPORT')


def report(self, msg, *args, **kwargs):
    """Log a message at the ``REPORT`` level."""
    self.log(LOG_LEVEL_REPORT, msg, *args, **kwargs)


setattr(logging, 'REPORT', LOG_LEVEL_REPORT)
setattr(logging.Logger, 'report', report)

# Convenience dictionary of available log level names and their log level integer
LOG_LEVELS = {
    logging.getLevelName(logging.NOTSET): logging.NOTSET,
    logging.getLevelName(logging.DEBUG): logging.DEBUG,
    logging.getLevelName(logging.INFO): logging.INFO,
    logging.getLevelName(LOG_LEVEL_REPORT): LOG_LEVEL_REPORT,
    logging.getLevelName(logging.WARNING): logging.WARNING,
    logging.getLevelName(logging.ERROR): logging.ERROR,
    logging.getLevelName(logging.CRITICAL): logging.CRITICAL,
}

AIIDA_LOGGER = logging.getLogger('aiida')
CLI_LOG_LEVEL = None


# The default logging dictionary for AiiDA that can be used in conjunction
# with the config.dictConfig method of python's logging module
def get_logging_config():
    from aiida.manage.configuration import get_config_option

    return {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'verbose': {
                'format': '%(levelname)s %(asctime)s %(module)s %(process)d '
                '%(thread)d %(message)s',
            },
            'halfverbose': {
                'format': '%(asctime)s <%(process)d> %(name)s: [%(levelname)s] %(message)s',
                'datefmt': '%m/%d/%Y %I:%M:%S %p',
            },
            'cli': {
                'class': 'aiida.cmdline.utils.log.CliFormatter'
            }
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'formatter': 'halfverbose',
            },
            'cli': {
                'class': 'aiida.cmdline.utils.log.CliHandler',
                'formatter': 'cli',
            }
        },
        'loggers': {
            'aiida': {
                'handlers': ['console'],
                'level': lambda: get_config_option('logging.aiida_loglevel'),
                'propagate': True,
            },
            'aiida.cmdline': {
                'handlers': ['cli'],
                'propagate': False,
            },
            'plumpy': {
                'handlers': ['console'],
                'level': lambda: get_config_option('logging.plumpy_loglevel'),
                'propagate': False,
            },
            'kiwipy': {
                'handlers': ['console'],
                'level': lambda: get_config_option('logging.kiwipy_loglevel'),
                'propagate': False,
            },
            'paramiko': {
                'handlers': ['console'],
                'level': lambda: get_config_option('logging.paramiko_loglevel'),
                'propagate': False,
            },
            'alembic': {
                'handlers': ['console'],
                'level': lambda: get_config_option('logging.alembic_loglevel'),
                'propagate': False,
            },
            'aio_pika': {
                'handlers': ['console'],
                'level': lambda: get_config_option('logging.aiopika_loglevel'),
                'propagate': False,
            },
            'sqlalchemy': {
                'handlers': ['console'],
                'level': lambda: get_config_option('logging.sqlalchemy_loglevel'),
                'propagate': False,
                'qualname': 'sqlalchemy.engine',
            },
            'py.warnings': {
                'handlers': ['console'],
            },
        },
    }


def evaluate_logging_configuration(dictionary):
    """Recursively evaluate the logging configuration, calling lambdas when encountered.

    This allows the configuration options that are dependent on the active profile to be loaded lazily.

    :return: evaluated logging configuration dictionary
    """
    result = {}

    for key, value in dictionary.items():
        if isinstance(value, collections.abc.Mapping):
            result[key] = evaluate_logging_configuration(value)
        elif isinstance(value, types.LambdaType):  # pylint: disable=no-member
            result[key] = value()
        else:
            result[key] = value

    return result


def configure_logging(with_orm=False, daemon=False, daemon_log_file=None):
    """
    Setup the logging by retrieving the LOGGING dictionary from aiida and passing it to
    the python module logging.config.dictConfig. If the logging needs to be setup for the
    daemon, set the argument 'daemon' to True and specify the path to the log file. This
    will cause a 'daemon_handler' to be added to all the configured loggers, that is a
    RotatingFileHandler that writes to the log file.

    :param with_orm: configure logging to the backend storage.
        We don't configure this by default, since it would load the modules that slow the CLI
    :param daemon: configure the logging for a daemon task by adding a file handler instead
        of the default 'console' StreamHandler
    :param daemon_log_file: absolute filepath of the log file for the RotatingFileHandler
    """
    from logging.config import dictConfig

    from aiida.manage.configuration import get_config_option

    # Evaluate the `LOGGING` configuration to resolve the lambdas that will retrieve the correct values based on the
    # currently configured profile.
    config = evaluate_logging_configuration(get_logging_config())
    daemon_handler_name = 'daemon_log_file'

    # Add the daemon file handler to all loggers if daemon=True
    if daemon is True:

        # Daemon always needs to run with ORM enabled
        with_orm = True

        if daemon_log_file is None:
            raise ValueError('daemon_log_file has to be defined when configuring for the daemon')

        config.setdefault('handlers', {})
        config['handlers'][daemon_handler_name] = {
            'level': 'DEBUG',
            'formatter': 'halfverbose',
            'class': 'logging.handlers.RotatingFileHandler',
            'filename': daemon_log_file,
            'encoding': 'utf8',
            'maxBytes': 10000000,  # 10 MB
            'backupCount': 10,
        }

        for logger in config.get('loggers', {}).values():
            logger.setdefault('handlers', []).append(daemon_handler_name)
            try:
                # Remove the `console` stdout stream handler to prevent messages being duplicated in the daemon log file
                logger['handlers'].remove('console')
            except ValueError:
                pass

    if CLI_LOG_LEVEL is not None:
        config['loggers']['aiida']['handlers'] = ['cli']
        config['loggers']['aiida']['level'] = CLI_LOG_LEVEL

    # Add the `DbLogHandler` if `with_orm` is `True`
    if with_orm:

        handler_dblogger = 'dblogger'

        config['handlers'][handler_dblogger] = {
            'level': get_config_option('logging.db_loglevel'),
            'class': 'aiida.orm.utils.log.DBLogHandler',
        }
        config['loggers']['aiida']['handlers'].append(handler_dblogger)

    dictConfig(config)


@contextlib.contextmanager
def override_log_level(level=logging.CRITICAL):
    """Temporarily adjust the log-level of logger."""
    logging.disable(level=level)
    try:
        yield
    finally:
        logging.disable(level=logging.NOTSET)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Provides functionality for managing large numbers of AiiDA Groups, via label delimitation."""
from collections import namedtuple
from functools import total_ordering
import re
from typing import Any, Iterator, List, Optional, Tuple
import warnings

from aiida import orm
from aiida.common.exceptions import NotExistent

__all__ = ('GroupPath', 'InvalidPath', 'GroupNotFoundError', 'GroupNotUniqueError', 'NoGroupsInPathError')

REGEX_ATTR = re.compile('^[a-zA-Z][\\_a-zA-Z0-9]*$')


class InvalidPath(Exception):
    """An exception to indicate that a path is not valid."""


class GroupNotFoundError(Exception):
    """An exception raised when a path does not have an associated group."""

    def __init__(self, grouppath):
        msg = f'No such group: {grouppath.path}'
        super().__init__(msg)


class GroupNotUniqueError(Exception):
    """An exception raised when a path has multiple associated groups."""

    def __init__(self, grouppath):
        msg = f'The path is not unique: {grouppath.path}'
        super().__init__(msg)


class NoGroupsInPathError(Exception):
    """An exception raised when a path has multiple associated groups."""

    def __init__(self, grouppath):
        msg = f'The path does not contain any descendant groups: {grouppath.path}'
        super().__init__(msg)


WalkNodeResult = namedtuple('WalkNodeResult', ['group_path', 'node'])


@total_ordering
class GroupPath:
    """A class to provide label delimited access to groups.

    See tests for usage examples.
    """

    def __init__(self, path: str = '', cls: orm.groups.GroupMeta = orm.Group, warn_invalid_child: bool = True) -> None:
        """Instantiate the class.

        :param path: The initial path of the group.
        :param cls: The subclass of `Group` to operate on.
        :param warn_invalid_child: Issue a warning, when iterating children, if a child path is invalid.

        """
        if not issubclass(cls, orm.Group):
            raise TypeError(f'cls must a subclass of Group: {cls}')

        self._delimiter = '/'
        self._cls = cls
        self._path_string = self._validate_path(path)
        self._path_list = self._path_string.split(self._delimiter) if path else []
        self._warn_invalid_child = warn_invalid_child

    def _validate_path(self, path):
        """Validate the supplied path."""
        if path == self._delimiter:
            return ''
        if self._delimiter * 2 in path:
            raise InvalidPath(f"The path may not contain a duplicate delimiter '{self._delimiter}': {path}")
        if (path.startswith(self._delimiter) or path.endswith(self._delimiter)):
            raise InvalidPath(f"The path may not start/end with the delimiter '{self._delimiter}': {path}")
        return path

    def __repr__(self) -> str:
        """Represent the instantiated class."""
        return f"{self.__class__.__name__}('{self.path}', cls='{self.cls}')"

    def __eq__(self, other: Any) -> bool:
        """Compare equality of path and ``Group`` subclass to another ``GroupPath`` object."""
        if not isinstance(other, GroupPath):
            return NotImplemented
        return (self.path, self.cls) == (other.path, other.cls)

    def __lt__(self, other: Any) -> bool:
        """Compare less-than operator of path and ``Group`` subclass to another ``GroupPath`` object."""
        if not isinstance(other, GroupPath):
            return NotImplemented
        return (self.path, self.cls) < (other.path, other.cls)

    @property
    def path(self) -> str:
        """Return the path string."""
        return self._path_string

    @property
    def path_list(self) -> List[str]:
        """Return a list of the path components."""
        return self._path_list[:]

    @property
    def key(self) -> Optional[str]:
        """Return the final component of the the path."""
        if self._path_list:
            return self._path_list[-1]
        return None

    @property
    def delimiter(self) -> str:
        """Return the delimiter used to split path into components."""
        return self._delimiter

    @property
    def cls(self) -> orm.groups.GroupMeta:
        """Return the cls used to query for and instantiate a ``Group`` with."""
        return self._cls

    @property
    def parent(self) -> Optional['GroupPath']:
        """Return the parent path."""
        if self.path_list:
            return GroupPath(
                self.delimiter.join(self.path_list[:-1]), cls=self.cls, warn_invalid_child=self._warn_invalid_child
            )
        return None

    def __truediv__(self, path: str) -> 'GroupPath':
        """Return a child ``GroupPath``, with a new path formed by appending ``path`` to the current path."""
        if not isinstance(path, str):
            raise TypeError(f'path is not a string: {path}')
        path = self._validate_path(path)
        child = GroupPath(
            path=self.path + self.delimiter + path if self.path else path,
            cls=self.cls,
            warn_invalid_child=self._warn_invalid_child
        )
        return child

    def __getitem__(self, path: str) -> 'GroupPath':
        """Return a child ``GroupPath``, with a new path formed by appending ``path`` to the current path."""
        return self.__truediv__(path)

    def get_group(self) -> Optional['GroupPath']:
        """Return the concrete group associated with this path."""
        try:
            return orm.QueryBuilder().append(self.cls, subclassing=False, filters={'label': self.path}).one()[0]
        except NotExistent:
            return None

    @property
    def group_ids(self) -> List[int]:
        """Return all the UUID associated with this GroupPath.

        :returns: and empty list, if no group associated with this label,
            or can be multiple if cls was None

        This is an efficient method for checking existence,
        which does not require the (slow) loading of the ORM entity.
        """
        query = orm.QueryBuilder()
        filters = {'label': self.path}
        query.append(self.cls, subclassing=False, filters=filters, project='id')
        return query.all(flat=True)

    @property
    def is_virtual(self) -> bool:
        """Return whether there is one or more concrete groups associated with this path."""
        return len(self.group_ids) == 0

    def get_or_create_group(self) -> Tuple[orm.Group, bool]:
        """Return the concrete group associated with this path or, create it, if it does not already exist."""
        return self.cls.collection.get_or_create(label=self.path)

    def delete_group(self):
        """Delete the concrete group associated with this path.

        :raises: GroupNotFoundError, GroupNotUniqueError
        """
        ids = self.group_ids
        if not ids:
            raise GroupNotFoundError(self)
        if len(ids) > 1:
            raise GroupNotUniqueError(self)
        self.cls.collection.delete(ids[0])

    @property
    def children(self) -> Iterator['GroupPath']:
        """Iterate through all (direct) children of this path."""
        query = orm.QueryBuilder()
        filters = {}
        if self.path:
            filters['label'] = {'like': f'{self.path + self.delimiter}%'}
        query.append(self.cls, subclassing=False, filters=filters, project='label')
        if query.count() == 0 and self.is_virtual:
            raise NoGroupsInPathError(self)

        yielded = []
        for (label,) in query.iterall():
            path = label.split(self._delimiter)
            if len(path) <= len(self._path_list):
                continue
            path_string = self._delimiter.join(path[:len(self._path_list) + 1])
            if (path_string not in yielded and path[:len(self._path_list)] == self._path_list):
                yielded.append(path_string)
                try:
                    yield GroupPath(path=path_string, cls=self.cls, warn_invalid_child=self._warn_invalid_child)
                except InvalidPath:
                    if self._warn_invalid_child:
                        warnings.warn(f'invalid path encountered: {path_string}')  # pylint: disable=no-member

    def __iter__(self) -> Iterator['GroupPath']:
        """Iterate through all (direct) children of this path."""
        return self.children

    def __len__(self) -> int:
        """Return the number of children for this path."""
        return sum(1 for _ in self.children)

    def __contains__(self, key: str) -> bool:
        """Return whether a child exists for this key."""
        for child in self.children:
            if child.path_list[-1] == key:
                return True
        return False

    def walk(self, return_virtual: bool = True) -> Iterator['GroupPath']:
        """Recursively iterate through all children of this path."""
        for child in self:
            if return_virtual or not child.is_virtual:
                yield child
            for sub_child in child.walk(return_virtual=return_virtual):
                if return_virtual or not sub_child.is_virtual:
                    yield sub_child

    def walk_nodes(
        self,
        filters: Optional[dict] = None,
        node_class: Optional[orm.Node] = None,
        query_batch: Optional[int] = None
    ) -> Iterator[WalkNodeResult]:
        """Recursively iterate through all nodes of this path and its children.

        :param filters: filters to apply to the node query
        :param node_class: return only nodes of a certain class (or list of classes)
        :param query_batch: The size of the batches to ask the backend to batch results in subcollections.
            You can optimize the speed of the query by tuning this parameter.
            Be aware though that is only safe if no commit will take place during this transaction.
        """
        query = orm.QueryBuilder()
        group_filters = {}
        if self.path:
            group_filters['label'] = {'or': [{'==': self.path}, {'like': f'{self.path + self.delimiter}%'}]}
        query.append(self.cls, subclassing=False, filters=group_filters, project='label', tag='group')
        query.append(
            orm.Node if node_class is None else node_class,
            with_group='group',
            filters=filters,
            project=['*'],
        )
        for (label, node) in query.iterall(query_batch) if query_batch else query.all():
            yield WalkNodeResult(GroupPath(label, cls=self.cls), node)

    @property
    def browse(self):
        """Return a ``GroupAttr`` instance, for attribute access to children."""
        return GroupAttr(self)


class GroupAttr:
    """A class to provide attribute access to a ``GroupPath`` children.

    The only public attributes on this class are dynamically created from the ``GroupPath`` child keys.
    NOTE: any child keys that do not conform to an acceptable (public) attribute string will be ignored.
    The ``GroupPath`` can be retrieved *via* a function call, e.g.::

        group_path = GroupPath()
        group_attr = GroupAttr(group_path)
        group_attr.a.b.c() == GroupPath("a/b/c")

    """

    def __init__(self, group_path: GroupPath) -> None:
        """Instantiate the ``GroupPath``, and a mapping of its children."""
        self._group_path = group_path

    def __repr__(self) -> str:
        """Represent the instantiated class."""
        return f"{self.__class__.__name__}('{self._group_path.path}', type='{self._group_path.cls}')"

    def __call__(self) -> GroupPath:
        """Return the ``GroupPath``."""
        return self._group_path

    def __dir__(self):
        """Return a list of available attributes."""
        return [c.path_list[-1] for c in self._group_path.children if REGEX_ATTR.match(c.path_list[-1])]

    def __getattr__(self, attr) -> 'GroupAttr':
        """Return the requested attribute name."""
        for child in self._group_path.children:
            if attr == child.path_list[-1]:
                return GroupAttr(child)
        raise AttributeError(attr)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Managing an AiiDA instance:

    * configuration file
    * profiles
    * databases
    * repositories
    * external components (such as Postgres, RabbitMQ)

.. note:: Modules in this sub package may require the database environment to be loaded

"""

# AUTO-GENERATED

# yapf: disable
# pylint: disable=wildcard-import

from .caching import *
from .configuration import *
from .external import *
from .manager import *

__all__ = (
    'BROKER_DEFAULTS',
    'CURRENT_CONFIG_VERSION',
    'CommunicationTimeout',
    'Config',
    'ConfigValidationError',
    'DEFAULT_DBINFO',
    'DeliveryFailed',
    'MIGRATIONS',
    'OLDEST_COMPATIBLE_CONFIG_VERSION',
    'Option',
    'Postgres',
    'PostgresConnectionMode',
    'ProcessLauncher',
    'Profile',
    'RemoteException',
    'check_and_migrate_config',
    'config_needs_migrating',
    'config_schema',
    'disable_caching',
    'downgrade_config',
    'enable_caching',
    'get_current_version',
    'get_manager',
    'get_option',
    'get_option_names',
    'get_use_cache',
    'parse_option',
    'upgrade_config',
)

# yapf: enable



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""AiiDA archive writer implementation."""
from datetime import datetime
import hashlib
from io import BytesIO
import json
import os
from pathlib import Path
import shutil
import tempfile
from typing import Any, BinaryIO, Dict, List, Literal, Optional, Set, Union
import zipfile

from archive_path import NOTSET, ZipPath, extract_file_in_zip, read_file_in_zip
from sqlalchemy import insert
from sqlalchemy.exc import IntegrityError as SqlaIntegrityError
from sqlalchemy.future.engine import Connection

from aiida import get_version
from aiida.common.exceptions import CorruptStorage, IncompatibleStorageSchema, IntegrityError
from aiida.common.hashing import chunked_file_hash
from aiida.common.progress_reporter import get_progress_reporter
from aiida.orm.entities import EntityTypes
from aiida.storage.sqlite_zip import models, utils
from aiida.tools.archive.abstract import ArchiveFormatAbstract, ArchiveWriterAbstract


class ArchiveWriterSqlZip(ArchiveWriterAbstract):
    """AiiDA archive writer implementation."""

    meta_name = utils.META_FILENAME
    db_name = utils.DB_FILENAME

    def __init__(
        self,
        path: Union[str, Path],
        fmt: ArchiveFormatAbstract,
        *,
        mode: Literal['x', 'w', 'a'] = 'x',
        compression: int = 6,
        work_dir: Optional[Path] = None,
        _debug: bool = False,
        _enforce_foreign_keys: bool = True,
    ):
        super().__init__(path, fmt, mode=mode, compression=compression)
        self._init_work_dir = work_dir
        self._in_context = False
        self._enforce_foreign_keys = _enforce_foreign_keys
        self._debug = _debug
        self._metadata: Dict[str, Any] = {}
        self._central_dir: Dict[str, Any] = {}
        self._deleted_paths: Set[str] = set()
        self._zip_path: Optional[ZipPath] = None
        self._work_dir: Optional[Path] = None
        self._conn: Optional[Connection] = None

    def _assert_in_context(self):
        if not self._in_context:
            raise AssertionError('Not in context')

    def __enter__(self) -> 'ArchiveWriterSqlZip':
        """Start writing to the archive"""
        self._metadata = {
            'export_version': self._format.latest_version,
            'aiida_version': get_version(),
            'key_format': 'sha256',
            'compression': self._compression,
        }
        self._work_dir = Path(tempfile.mkdtemp()) if self._init_work_dir is None else Path(self._init_work_dir)
        self._central_dir = {}
        self._zip_path = ZipPath(
            self._path,
            mode=self._mode,
            compression=zipfile.ZIP_DEFLATED if self._compression else zipfile.ZIP_STORED,
            compresslevel=self._compression,
            info_order=(self.meta_name, self.db_name),
            name_to_info=self._central_dir,
        )
        engine = utils.create_sqla_engine(
            self._work_dir / self.db_name, enforce_foreign_keys=self._enforce_foreign_keys, echo=self._debug
        )
        models.SqliteBase.metadata.create_all(engine)
        self._conn = engine.connect()
        self._in_context = True
        return self

    def __exit__(self, *args, **kwargs):
        """Finalise the archive"""
        if self._conn:
            self._conn.commit()
            self._conn.close()
        assert self._work_dir is not None
        with (self._work_dir / self.db_name).open('rb') as handle:
            self._stream_binary(self.db_name, handle)
        self._stream_binary(
            self.meta_name,
            BytesIO(json.dumps(self._metadata).encode('utf8')),
            compression=0,  # the metadata is small, so no benefit for compression
        )
        if self._zip_path:
            self._zip_path.close()
            self._central_dir = {}
        if self._work_dir is not None and self._init_work_dir is None:
            shutil.rmtree(self._work_dir, ignore_errors=True)
        self._zip_path = self._work_dir = self._conn = None
        self._in_context = False

    def update_metadata(self, data: Dict[str, Any], overwrite: bool = False) -> None:
        if not overwrite and set(self._metadata).intersection(set(data)):
            raise ValueError(f'Cannot overwrite existing keys: {set(self._metadata).intersection(set(data))}')
        self._metadata.update(data)

    def bulk_insert(
        self,
        entity_type: EntityTypes,
        rows: List[Dict[str, Any]],
        allow_defaults: bool = False,
    ) -> None:
        if not rows:
            return
        self._assert_in_context()
        assert self._conn is not None
        model, col_keys = models.get_model_from_entity(entity_type)
        if allow_defaults:
            for row in rows:
                if not col_keys.issuperset(row):
                    raise IntegrityError(
                        f'Incorrect fields given for {entity_type}: {set(row)} not subset of {col_keys}'
                    )
        else:
            for row in rows:
                if set(row) != col_keys:
                    raise IntegrityError(f'Incorrect fields given for {entity_type}: {set(row)} != {col_keys}')
        try:
            self._conn.execute(insert(model.__table__), rows)
        except SqlaIntegrityError as exc:
            raise IntegrityError(f'Inserting {entity_type}: {exc}') from exc

    def _stream_binary(
        self,
        name: str,
        handle: BinaryIO,
        *,
        buffer_size: Optional[int] = None,
        compression: Optional[int] = None,
        comment: Optional[bytes] = None,
    ) -> None:
        """Add a binary stream to the archive.

        :param buffer_size: Number of bytes to buffer
        :param compression: Override global compression level
        :param comment: A binary meta comment about the object
        """
        self._assert_in_context()
        assert self._zip_path is not None
        kwargs: Dict[str, Any] = {'comment': NOTSET if comment is None else comment}
        if compression is not None:
            kwargs['compression'] = zipfile.ZIP_DEFLATED if compression else zipfile.ZIP_STORED
            kwargs['level'] = compression

        # compute the file size of the handle
        position = handle.tell()
        handle.seek(0, os.SEEK_END)
        file_size = handle.tell()
        handle.seek(position)

        with self._zip_path.joinpath(name).open(mode='wb', file_size=file_size, **kwargs) as zip_handle:
            if buffer_size is None:
                shutil.copyfileobj(handle, zip_handle)
            else:
                shutil.copyfileobj(handle, zip_handle, length=buffer_size)

    def put_object(self, stream: BinaryIO, *, buffer_size: Optional[int] = None, key: Optional[str] = None) -> str:
        if key is None:
            key = chunked_file_hash(stream, hashlib.sha256)
            stream.seek(0)
        if f'{utils.REPO_FOLDER}/{key}' not in self._central_dir:
            self._stream_binary(f'{utils.REPO_FOLDER}/{key}', stream, buffer_size=buffer_size)
        return key

    def delete_object(self, key: str) -> None:
        raise IOError(f'Cannot delete objects in {self._mode!r} mode')


class ArchiveAppenderSqlZip(ArchiveWriterSqlZip):
    """AiiDA archive appender implementation."""

    def delete_object(self, key: str) -> None:
        self._assert_in_context()
        if f'{utils.REPO_FOLDER}/{key}' in self._central_dir:
            raise IOError(f'Cannot delete object {key!r} that has been added in the same append context')
        self._deleted_paths.add(f'{utils.REPO_FOLDER}/{key}')

    def __enter__(self) -> 'ArchiveAppenderSqlZip':
        """Start appending to the archive"""
        # the file should already exist
        if not self._path.exists():
            raise FileNotFoundError(f'Archive {self._path} does not exist')
        # the file should be an archive with the correct version
        version = self._format.read_version(self._path)
        if not version == self._format.latest_version:
            raise IncompatibleStorageSchema(
                f'Archive is version {version!r} but expected {self._format.latest_version!r}'
            )
        # load the metadata
        self._metadata = json.loads(read_file_in_zip(self._path, utils.META_FILENAME, 'utf8', search_limit=4))
        # overwrite metadata
        self._metadata['mtime'] = datetime.now().isoformat()
        self._metadata['compression'] = self._compression
        # create the work folder
        self._work_dir = Path(tempfile.mkdtemp()) if self._init_work_dir is None else Path(self._init_work_dir)
        # create a new zip file in the work folder
        self._central_dir = {}
        self._deleted_paths = set()
        self._zip_path = ZipPath(
            self._work_dir / 'archive.zip',
            mode='w',
            compression=zipfile.ZIP_DEFLATED if self._compression else zipfile.ZIP_STORED,
            compresslevel=self._compression,
            info_order=(self.meta_name, self.db_name),
            name_to_info=self._central_dir,
        )
        # extract the database to the work folder
        db_file = self._work_dir / self.db_name
        with db_file.open('wb') as handle:
            try:
                extract_file_in_zip(self.path, utils.DB_FILENAME, handle, search_limit=4)
            except Exception as exc:
                raise CorruptStorage(f'archive database could not be read: {exc}') from exc
        # open a connection to the database
        engine = utils.create_sqla_engine(
            self._work_dir / self.db_name, enforce_foreign_keys=self._enforce_foreign_keys, echo=self._debug
        )
        # to-do could check that the database has correct schema:
        # https://docs.sqlalchemy.org/en/14/core/reflection.html#reflecting-all-tables-at-once
        self._conn = engine.connect()
        self._in_context = True
        return self

    def __exit__(self, *args, **kwargs):
        """Finalise the archive"""
        if self._conn:
            self._conn.commit()
            self._conn.close()
        assert self._work_dir is not None
        # write the database and metadata to the new archive
        with (self._work_dir / self.db_name).open('rb') as handle:
            self._stream_binary(self.db_name, handle)
        self._stream_binary(
            self.meta_name,
            BytesIO(json.dumps(self._metadata).encode('utf8')),
            compression=0,
        )
        # finalise the new archive
        self._copy_old_zip_files()
        if self._zip_path is not None:
            self._zip_path.close()
        self._central_dir = {}
        self._deleted_paths = set()
        # now move it to the original location
        self._path.unlink()
        shutil.move(self._work_dir / 'archive.zip', self._path)  # type: ignore[arg-type]
        if self._init_work_dir is None:
            shutil.rmtree(self._work_dir, ignore_errors=True)
        self._zip_path = self._work_dir = self._conn = None
        self._in_context = False

    def _copy_old_zip_files(self):
        """Copy the old archive content to the new one (omitting any amended or deleted files)"""
        assert self._zip_path is not None
        with ZipPath(self._path, mode='r') as old_archive:
            length = sum(1 for _ in old_archive.glob('**/*', include_virtual=False))
            with get_progress_reporter()(desc='Writing amended archive', total=length) as progress:
                for subpath in old_archive.glob('**/*', include_virtual=False):
                    if subpath.at in self._central_dir or subpath.at in self._deleted_paths:
                        continue
                    new_path_sub = self._zip_path.joinpath(subpath.at)
                    if subpath.is_dir():
                        new_path_sub.mkdir(exist_ok=True)
                    else:
                        new_path_sub.putfile(subpath)
                    progress.update()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=import-error,no-name-in-module
"""Module to manage nodes for the SQLA backend."""
from sqlalchemy.dialects.postgresql import JSONB, UUID
from sqlalchemy.orm import backref, relationship
from sqlalchemy.schema import Column
from sqlalchemy.sql.schema import ForeignKey, Index
from sqlalchemy.types import DateTime, Integer, String, Text

from aiida.common import timezone
from aiida.common.utils import get_new_uuid
from aiida.storage.psql_dos.models.base import Base


class DbNode(Base):
    """Database model to store data for :py:class:`aiida.orm.Node`.

    Each node can be categorized according to its ``node_type``,
    which indicates what kind of data or process node it is.
    Additionally, process nodes also have a ``process_type`` that further indicates what is the specific plugin it uses.

    Nodes can also store two kind of properties:

    - ``attributes`` are determined by the ``node_type``,
      and are set before storing the node and can't be modified afterwards.
    - ``extras``, on the other hand,
      can be added and removed after the node has been stored and are usually set by the user.

    """

    __tablename__ = 'db_dbnode'

    id = Column(Integer, primary_key=True)  # pylint: disable=invalid-name
    uuid = Column(UUID(as_uuid=True), default=get_new_uuid, nullable=False, unique=True)
    node_type = Column(String(255), default='', nullable=False, index=True)
    process_type = Column(String(255), index=True)
    label = Column(String(255), nullable=False, default='', index=True)
    description = Column(Text(), nullable=False, default='')
    ctime = Column(DateTime(timezone=True), default=timezone.now, nullable=False, index=True)
    mtime = Column(DateTime(timezone=True), default=timezone.now, onupdate=timezone.now, nullable=False, index=True)
    attributes = Column(JSONB, default=dict)
    extras = Column(JSONB, default=dict)
    repository_metadata = Column(JSONB, nullable=False, default=dict)
    dbcomputer_id = Column(
        Integer,
        ForeignKey('db_dbcomputer.id', deferrable=True, initially='DEFERRED', ondelete='RESTRICT'),
        nullable=True,
        index=True
    )
    user_id = Column(
        Integer,
        ForeignKey('db_dbuser.id', deferrable=True, initially='DEFERRED', ondelete='RESTRICT'),
        nullable=False,
        index=True
    )

    # pylint: disable=fixme
    # TODO SP: The 'passive_deletes=all' argument here means that SQLAlchemy
    # won't take care of automatic deleting in the DbLink table. This still
    # isn't exactly the same behaviour than with Django. The solution to
    # this is probably a ON DELETE inside the DB. On removing node with id=x,
    # we would remove all link with x as an output.

    dbcomputer = relationship('DbComputer', backref=backref('dbnodes', passive_deletes='all', cascade='merge'))
    user = relationship('DbUser', backref=backref(
        'dbnodes',
        passive_deletes='all',
        cascade='merge',
    ))

    # outputs via db_dblink table
    outputs_q = relationship(
        'DbNode',
        secondary='db_dblink',
        primaryjoin='DbNode.id == DbLink.input_id',
        secondaryjoin='DbNode.id == DbLink.output_id',
        backref=backref('inputs_q', passive_deletes=True, lazy='dynamic'),
        lazy='dynamic',
        passive_deletes=True
    )

    __table_args__ = (
        Index(
            'ix_pat_db_dbnode_label', label, postgresql_using='btree', postgresql_ops={'label': 'varchar_pattern_ops'}
        ),
        Index(
            'ix_pat_db_dbnode_node_type',
            node_type,
            postgresql_using='btree',
            postgresql_ops={'node_type': 'varchar_pattern_ops'}
        ),
        Index(
            'ix_pat_db_dbnode_process_type',
            process_type,
            postgresql_using='btree',
            postgresql_ops={'process_type': 'varchar_pattern_ops'}
        ),
    )

    @property
    def outputs(self):
        return self.outputs_q.all()

    @property
    def inputs(self):
        return self.inputs_q.all()  # pylint: disable=no-member

    def get_simple_name(self, invalid_result=None):
        """
        Return a string with the last part of the type name.

        If the type is empty, use 'Node'.
        If the type is invalid, return the content of the input variable
        ``invalid_result``.

        :param invalid_result: The value to be returned if the node type is
            not recognized.
        """
        thistype = self.node_type
        # Fix for base class
        if thistype == '':
            thistype = 'node.Node.'
        if not thistype.endswith('.'):
            return invalid_result
        thistype = thistype[:-1]  # Strip final dot
        return thistype.rpartition('.')[2]

    @property
    def pk(self):
        return self.id

    def __str__(self):
        """Get string object out of DbNode object."""
        simplename = self.get_simple_name(invalid_result='Unknown')
        # node pk + type
        if self.label:
            return f'{simplename} node [{self.pk}]: {self.label}'
        return f'{simplename} node [{self.pk}]'


class DbLink(Base):
    """Database model to store links between :py:class:`aiida.orm.Node`.

    Each entry in this table contains not only the ``id`` information of the two nodes that are linked,
    but also some extra properties of the link themselves.
    This includes the ``type`` of the link (see the :ref:`topics:provenance:concepts` section for all possible types)
    as well as a ``label`` which is more specific and typically determined by
    the procedure generating the process node that links the data nodes.
    """

    __tablename__ = 'db_dblink'

    id = Column(Integer, primary_key=True)  # pylint: disable=invalid-name
    input_id = Column(
        Integer, ForeignKey('db_dbnode.id', deferrable=True, initially='DEFERRED'), nullable=False, index=True
    )
    output_id = Column(
        Integer,
        ForeignKey('db_dbnode.id', ondelete='CASCADE', deferrable=True, initially='DEFERRED'),
        nullable=False,
        index=True
    )

    # https://docs.sqlalchemy.org/en/14/errors.html#relationship-x-will-copy-column-q-to-column-p-which-conflicts-with-relationship-s-y
    input = relationship('DbNode', primaryjoin='DbLink.input_id == DbNode.id', overlaps='inputs_q,outputs_q')
    output = relationship('DbNode', primaryjoin='DbLink.output_id == DbNode.id', overlaps='inputs_q,outputs_q')

    label = Column(String(255), nullable=False, index=True)
    type = Column(String(255), nullable=False, index=True)

    # A calculation can have both a 'return' and a 'create' link to
    # a single data output node, which would violate the unique constraint
    # defined below, since the difference in link type is not considered.
    # The distinction between the type of a 'create' and a 'return' link is not
    # implemented at the moment, so the unique constraint is disabled.
    __table_args__ = (
        # I cannot add twice the same link
        # I want unique labels among all inputs of a node
        # UniqueConstraint('output_id', 'label'),
        Index(
            'ix_pat_db_dblink_label', label, postgresql_using='btree', postgresql_ops={'label': 'varchar_pattern_ops'}
        ),
        Index('ix_pat_db_dblink_type', type, postgresql_using='btree', postgresql_ops={'type': 'varchar_pattern_ops'}),
    )

    def __str__(self):
        return '{} ({}) --> {} ({})'.format(
            self.input.get_simple_name(invalid_result='Unknown node'), self.input.pk,
            self.output.get_simple_name(invalid_result='Unknown node'), self.output.pk
        )



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name
"""Invalidating node hash

Users should rehash nodes for caching

Revision ID: e797afa09270
Revises: 26d561acd560
Create Date: 2019-07-01 19:39:33.605457

"""
from alembic import op

from aiida.storage.psql_dos.migrations.utils.integrity import drop_hashes

# revision identifiers, used by Alembic.
revision = 'e797afa09270'
down_revision = '26d561acd560'
branch_labels = None
depends_on = None


def upgrade():
    """drop the hashes when upgrading"""
    drop_hashes(op.get_bind())  # pylint: disable=no-member


def downgrade():
    """drop the hashes also when downgrading"""
    drop_hashes(op.get_bind())  # pylint: disable=no-member



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Components for the WorkChain concept of the workflow engine."""
import collections.abc
import functools
import logging
from typing import TYPE_CHECKING, Any, List, Optional, Sequence, Tuple, Union

from plumpy.persistence import auto_persist
from plumpy.process_states import Continue, Wait
from plumpy.workchains import Stepper
from plumpy.workchains import WorkChainSpec as PlumpyWorkChainSpec
from plumpy.workchains import _PropagateReturn, if_, return_, while_

from aiida.common import exceptions
from aiida.common.extendeddicts import AttributeDict
from aiida.common.lang import override
from aiida.orm import Node, ProcessNode, WorkChainNode
from aiida.orm.utils import load_node

from ..exit_code import ExitCode
from ..process import Process, ProcessState
from ..process_spec import ProcessSpec
from .awaitable import Awaitable, AwaitableAction, AwaitableTarget, construct_awaitable

if TYPE_CHECKING:
    from aiida.engine.runners import Runner

__all__ = ('WorkChain', 'if_', 'while_', 'return_')


class WorkChainSpec(ProcessSpec, PlumpyWorkChainSpec):
    pass


@auto_persist('_awaitables')
class WorkChain(Process):
    """The `WorkChain` class is the principle component to implement workflows in AiiDA."""

    _node_class = WorkChainNode
    _spec_class = WorkChainSpec
    _STEPPER_STATE = 'stepper_state'
    _CONTEXT = 'CONTEXT'

    def __init__(
        self,
        inputs: Optional[dict] = None,
        logger: Optional[logging.Logger] = None,
        runner: Optional['Runner'] = None,
        enable_persistence: bool = True
    ) -> None:
        """Construct a WorkChain instance.

        Construct the instance only if it is a sub class of `WorkChain`, otherwise raise `InvalidOperation`.

        :param inputs: work chain inputs
        :param logger: aiida logger
        :param runner: work chain runner
        :param enable_persistence: whether to persist this work chain

        """
        if self.__class__ == WorkChain:
            raise exceptions.InvalidOperation('cannot construct or launch a base `WorkChain` class.')

        super().__init__(inputs, logger, runner, enable_persistence=enable_persistence)

        self._stepper: Optional[Stepper] = None
        self._awaitables: List[Awaitable] = []
        self._context = AttributeDict()

    @classmethod
    def spec(cls) -> WorkChainSpec:
        return super().spec()  # type: ignore[return-value]

    @property
    def ctx(self) -> AttributeDict:
        """Get the context."""
        return self._context

    @override
    def save_instance_state(self, out_state, save_context):
        """Save instance state.

        :param out_state: state to save in

        :param save_context:
        :type save_context: :class:`!plumpy.persistence.LoadSaveContext`

        """
        super().save_instance_state(out_state, save_context)
        # Save the context
        out_state[self._CONTEXT] = self.ctx

        # Ask the stepper to save itself
        if self._stepper is not None:
            out_state[self._STEPPER_STATE] = self._stepper.save()

    @override
    def load_instance_state(self, saved_state, load_context):
        super().load_instance_state(saved_state, load_context)
        # Load the context
        self._context = saved_state[self._CONTEXT]

        # Recreate the stepper
        self._stepper = None
        stepper_state = saved_state.get(self._STEPPER_STATE, None)
        if stepper_state is not None:
            self._stepper = self.spec().get_outline().recreate_stepper(stepper_state, self)  # type: ignore[arg-type]

        self.set_logger(self.node.logger)

        if self._awaitables:
            self.action_awaitables()

    def on_run(self):
        super().on_run()
        self.node.set_stepper_state_info(str(self._stepper))

    def _resolve_nested_context(self, key: str) -> Tuple[AttributeDict, str]:
        """
        Returns a reference to a sub-dictionary of the context and the last key,
        after resolving a potentially segmented key where required sub-dictionaries are created as needed.

        :param key: A key into the context, where words before a dot are interpreted as a key for a sub-dictionary
        """
        ctx = self.ctx
        ctx_path = key.split('.')

        for index, path in enumerate(ctx_path[:-1]):
            try:
                ctx = ctx[path]
            except KeyError:  # see below why this is the only exception we have to catch here
                ctx[path] = AttributeDict()  # create the sub-dict and update the context
                ctx = ctx[path]
                continue

            # Notes:
            # * the first ctx (self.ctx) is guaranteed to be an AttributeDict, hence the post-"dereference" checking
            # * the values can be many different things: on insertion they are either AtrributeDict, List or Awaitables
            #   (subclasses of AttributeDict) but after resolution of an Awaitable this will be the value itself
            # * assumption: a resolved value is never a plain AttributeDict, on the other hand if a resolved Awaitable
            #   would be an AttributeDict we can append things to it since the order of tasks is maintained.
            if type(ctx) != AttributeDict:  # pylint: disable=C0123
                raise ValueError(
                    f'Can not update the context for key `{key}`:'
                    f' found instance of `{type(ctx)}` at `{".".join(ctx_path[:index+1])}`, expected AttributeDict'
                )

        return ctx, ctx_path[-1]

    def insert_awaitable(self, awaitable: Awaitable) -> None:
        """Insert an awaitable that should be terminated before before continuing to the next step.

        :param awaitable: the thing to await
        """
        ctx, key = self._resolve_nested_context(awaitable.key)

        # Already assign the awaitable itself to the location in the context container where it is supposed to end up
        # once it is resolved. This is especially important for the `APPEND` action, since it needs to maintain the
        # order, but the awaitables will not necessarily be resolved in the order in which they are added. By using the
        # awaitable as a placeholder, in the `resolve_awaitable`, it can be found and replaced by the resolved value.
        if awaitable.action == AwaitableAction.ASSIGN:
            ctx[key] = awaitable
        elif awaitable.action == AwaitableAction.APPEND:
            ctx.setdefault(key, []).append(awaitable)
        else:
            raise AssertionError(f'Unsupported awaitable action: {awaitable.action}')

        self._awaitables.append(
            awaitable
        )  # add only if everything went ok, otherwise we end up in an inconsistent state
        self._update_process_status()

    def resolve_awaitable(self, awaitable: Awaitable, value: Any) -> None:
        """Resolve an awaitable.

        Precondition: must be an awaitable that was previously inserted.

        :param awaitable: the awaitable to resolve
        """

        ctx, key = self._resolve_nested_context(awaitable.key)

        if awaitable.action == AwaitableAction.ASSIGN:
            ctx[key] = value
        elif awaitable.action == AwaitableAction.APPEND:
            # Find the same awaitable inserted in the context
            container = ctx[key]
            for index, placeholder in enumerate(container):
                if isinstance(placeholder, Awaitable) and placeholder.pk == awaitable.pk:
                    container[index] = value
                    break
            else:
                raise AssertionError(f'Awaitable `{awaitable.pk} was not found in `ctx.{awaitable.key}`')
        else:
            raise AssertionError(f'Unsupported awaitable action: {awaitable.action}')

        awaitable.resolved = True
        self._awaitables.remove(awaitable)  # remove only if everything went ok, otherwise we may lose track

        if not self.has_terminated():
            # the process may be terminated, for example, if the process was killed or excepted
            # then we should not try to update it
            self._update_process_status()

    def to_context(self, **kwargs: Union[Awaitable, ProcessNode]) -> None:
        """Add a dictionary of awaitables to the context.

        This is a convenience method that provides syntactic sugar, for a user to add multiple intersteps that will
        assign a certain value to the corresponding key in the context of the work chain.
        """
        for key, value in kwargs.items():
            awaitable = construct_awaitable(value)
            awaitable.key = key
            self.insert_awaitable(awaitable)

    def _update_process_status(self) -> None:
        """Set the process status with a message accounting the current sub processes that we are waiting for."""
        if self._awaitables:
            status = f"Waiting for child processes: {', '.join([str(_.pk) for _ in self._awaitables])}"
            self.node.set_process_status(status)
        else:
            self.node.set_process_status(None)

    @override
    def run(self) -> Any:
        self._stepper = self.spec().get_outline().create_stepper(self)  # type: ignore[arg-type]
        return self._do_step()

    def _do_step(self) -> Any:
        """Execute the next step in the outline and return the result.

        If the stepper returns a non-finished status and the return value is of type ToContext, the contents of the
        ToContext container will be turned into awaitables if necessary. If any awaitables were created, the process
        will enter in the Wait state, otherwise it will go to Continue. When the stepper returns that it is done, the
        stepper result will be converted to None and returned, unless it is an integer or instance of ExitCode.
        """
        from .context import ToContext

        self._awaitables = []
        result: Any = None

        try:
            assert self._stepper is not None
            finished, stepper_result = self._stepper.step()
        except _PropagateReturn as exception:
            finished, result = True, exception.exit_code
        else:
            # Set result to None unless stepper_result was non-zero positive integer or ExitCode with similar status
            if isinstance(stepper_result, int) and stepper_result > 0:
                result = ExitCode(stepper_result)
            elif isinstance(stepper_result, ExitCode) and stepper_result.status > 0:
                result = stepper_result
            else:
                result = None

        # If the stepper said we are finished or the result is an ExitCode, we exit by returning
        if finished or isinstance(result, ExitCode):
            return result

        if isinstance(stepper_result, ToContext):
            self.to_context(**stepper_result)

        if self._awaitables:
            return Wait(self._do_step, 'Waiting before next step')

        return Continue(self._do_step)

    def _store_nodes(self, data: Any) -> None:
        """Recurse through a data structure and store any unstored nodes that are found along the way

        :param data: a data structure potentially containing unstored nodes
        """
        if isinstance(data, Node) and not data.is_stored:
            data.store()
        elif isinstance(data, collections.abc.Mapping):
            for _, value in data.items():
                self._store_nodes(value)
        elif isinstance(data, collections.abc.Sequence) and not isinstance(data, str):
            for value in data:
                self._store_nodes(value)

    @override
    def on_exiting(self) -> None:
        """Ensure that any unstored nodes in the context are stored, before the state is exited

        After the state is exited the next state will be entered and if persistence is enabled, a checkpoint will
        be saved. If the context contains unstored nodes, the serialization necessary for checkpointing will fail.
        """
        super().on_exiting()
        try:
            self._store_nodes(self.ctx)
        except Exception:  # pylint: disable=broad-except
            # An uncaught exception here will have bizarre and disastrous consequences
            self.logger.exception('exception in _store_nodes called in on_exiting')

    def on_wait(self, awaitables: Sequence[Awaitable]):
        """Entering the WAITING state."""
        super().on_wait(awaitables)
        if self._awaitables:
            self.action_awaitables()
        else:
            self.call_soon(self.resume)

    def action_awaitables(self) -> None:
        """Handle the awaitables that are currently registered with the work chain.

        Depending on the class type of the awaitable's target a different callback
        function will be bound with the awaitable and the runner will be asked to
        call it when the target is completed
        """
        for awaitable in self._awaitables:
            if awaitable.target == AwaitableTarget.PROCESS:
                callback = functools.partial(self.call_soon, self.on_process_finished, awaitable)
                self.runner.call_on_process_finish(awaitable.pk, callback)
            else:
                assert f"invalid awaitable target '{awaitable.target}'"

    def on_process_finished(self, awaitable: Awaitable) -> None:
        """Callback function called by the runner when the process instance identified by pk is completed.

        The awaitable will be effectuated on the context of the work chain and removed from the internal list. If all
        awaitables have been dealt with, the work chain process is resumed.

        :param awaitable: an Awaitable instance
        """
        self.logger.info('received callback that awaitable %d has terminated', awaitable.pk)

        try:
            node = load_node(awaitable.pk)
        except (exceptions.MultipleObjectsError, exceptions.NotExistent):
            raise ValueError(f'provided pk<{awaitable.pk}> could not be resolved to a valid Node instance')

        if awaitable.outputs:
            value = {entry.link_label: entry.node for entry in node.base.links.get_outgoing()}
        else:
            value = node

        self.resolve_awaitable(awaitable, value)

        if self.state == ProcessState.WAITING and not self._awaitables:
            self.resume()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module that define the exceptions that are thrown by AiiDA's internal code."""

__all__ = (
    'AiidaException', 'NotExistent', 'NotExistentAttributeError', 'NotExistentKeyError', 'MultipleObjectsError',
    'RemoteOperationError', 'ContentNotExistent', 'FailedError', 'StoringNotAllowed', 'ModificationNotAllowed',
    'IntegrityError', 'UniquenessError', 'EntryPointError', 'MissingEntryPointError', 'MultipleEntryPointError',
    'LoadingEntryPointError', 'InvalidEntryPointTypeError', 'InvalidOperation', 'ParsingError', 'InternalError',
    'PluginInternalError', 'ValidationError', 'ConfigurationError', 'ProfileConfigurationError',
    'MissingConfigurationError', 'ConfigurationVersionError', 'IncompatibleStorageSchema', 'CorruptStorage',
    'DbContentError', 'InputValidationError', 'FeatureNotAvailable', 'FeatureDisabled', 'LicensingException',
    'TestsNotAllowedError', 'UnsupportedSpeciesError', 'TransportTaskException', 'OutputParsingError', 'HashingError',
    'StorageMigrationError', 'LockedProfileError', 'LockingProfileError', 'ClosedStorage'
)


class AiidaException(Exception):
    """
    Base class for all AiiDA exceptions.

    Each module will have its own subclass, inherited from this
    (e.g. ExecManagerException, TransportException, ...)
    """


class NotExistent(AiidaException):
    """
    Raised when the required entity does not exist.
    """


class NotExistentAttributeError(AttributeError, NotExistent):
    """
    Raised when the required entity does not exist, when fetched as an attribute.
    """


class NotExistentKeyError(KeyError, NotExistent):
    """
    Raised when the required entity does not exist, when fetched as a dictionary key.
    """


class MultipleObjectsError(AiidaException):
    """
    Raised when more than one entity is found in the DB, but only one was
    expected.
    """


class RemoteOperationError(AiidaException):
    """
    Raised when an error in a remote operation occurs, as in a failed kill()
    of a scheduler job.
    """


class ContentNotExistent(NotExistent):
    """
    Raised when trying to access an attribute, a key or a file in the result
    nodes that is not present
    """


class FailedError(AiidaException):
    """
    Raised when accessing a calculation that is in the FAILED status
    """


class StoringNotAllowed(AiidaException):
    """
    Raised when the user tries to store an unstorable node (e.g. a base Node class)
    """


class ModificationNotAllowed(AiidaException):
    """
    Raised when the user tries to modify a field, object, property, ... that should not
    be modified.
    """


class IntegrityError(AiidaException):
    """
    Raised when there is an underlying data integrity error.  This can be database related
    or a general data integrity error.  This can happen if, e.g., a foreign key check fails.
    See PEP 249 for details.
    """


class UniquenessError(AiidaException):
    """
    Raised when the user tries to violate a uniqueness constraint (on the
    DB, for instance).
    """


class EntryPointError(AiidaException):
    """Raised when an entry point cannot be uniquely resolved and imported."""


class MissingEntryPointError(EntryPointError):
    """Raised when the requested entry point is not registered with the entry point manager."""


class MultipleEntryPointError(EntryPointError):
    """Raised when the requested entry point cannot uniquely be resolved by the entry point manager."""


class LoadingEntryPointError(EntryPointError):
    """Raised when the resource corresponding to requested entry point cannot be imported."""


class InvalidEntryPointTypeError(EntryPointError):
    """Raised when a loaded entry point has a type that is not supported by the corresponding entry point group."""


class InvalidOperation(AiidaException):
    """
    The allowed operation is not valid (e.g., when trying to add a non-internal attribute
    before saving the entry), or deleting an entry that is protected (e.g.,
    because it is referenced by foreign keys)
    """


class ParsingError(AiidaException):
    """
    Generic error raised when there is a parsing error
    """


class InternalError(AiidaException):
    """
    Error raised when there is an internal error of AiiDA.
    """


class PluginInternalError(InternalError):
    """
    Error raised when there is an internal error which is due to a plugin
    and not to the AiiDA infrastructure.
    """


class ValidationError(AiidaException):
    """
    Error raised when there is an error during the validation phase
    of a property.
    """


class ConfigurationError(AiidaException):
    """
    Error raised when there is a configuration error in AiiDA.
    """


class ProfileConfigurationError(ConfigurationError):
    """
    Configuration error raised when a wrong/inexistent profile is requested.
    """


class MissingConfigurationError(ConfigurationError):
    """
    Configuration error raised when the configuration file is missing.
    """


class ConfigurationVersionError(ConfigurationError):
    """
    Configuration error raised when the configuration file version is not
    compatible with the current version.
    """


class ClosedStorage(AiidaException):
    """Raised when trying to access data from a closed storage backend."""


class UnreachableStorage(ConfigurationError):
    """Raised when a connection to the storage backend fails."""


class IncompatibleDatabaseSchema(ConfigurationError):
    """Raised when the storage schema is incompatible with that of the code.

    Deprecated for ``IncompatibleStorageSchema``
    """


class IncompatibleStorageSchema(IncompatibleDatabaseSchema):
    """Raised when the storage schema is incompatible with that of the code."""


class CorruptStorage(ConfigurationError):
    """Raised when the storage is not found to be internally consistent on validation."""


class DatabaseMigrationError(AiidaException):
    """Raised if a critical error is encountered during a storage migration.

    Deprecated for ``StorageMigrationError``
    """


class StorageMigrationError(DatabaseMigrationError):
    """Raised if a critical error is encountered during a storage migration."""


class DbContentError(AiidaException):
    """
    Raised when the content of the DB is not valid.
    This should never happen if the user does not play directly
    with the DB.
    """


class InputValidationError(ValidationError):
    """
    The input data for a calculation did not validate (e.g., missing
    required input data, wrong data, ...)
    """


class FeatureNotAvailable(AiidaException):
    """
    Raised when a feature is requested from a plugin, that is not available.
    """


class FeatureDisabled(AiidaException):
    """
    Raised when a feature is requested, but the user has chosen to disable
    it (e.g., for submissions on disabled computers).
    """


class LicensingException(AiidaException):
    """
    Raised when requirements for data licensing are not met.
    """


class TestsNotAllowedError(AiidaException):
    """
    Raised when tests are required to be run/loaded, but we are not in a testing environment.

    This is to prevent data loss.
    """


class UnsupportedSpeciesError(ValueError):
    """
    Raised when StructureData operations are fed species that are not supported by AiiDA such as Deuterium
    """


class TransportTaskException(AiidaException):
    """
    Raised when a TransportTask, an task to be completed by the engine that requires transport, fails
    """


class OutputParsingError(ParsingError):
    """
    Can be raised by a Parser when it fails to parse the output generated by a `CalcJob` process.
    """


class CircusCallError(AiidaException):
    """
    Raised when an attempt to contact Circus returns an error in the response
    """


class HashingError(AiidaException):
    """
    Raised when an attempt to hash an object fails via a known failure mode
    """


class LockedProfileError(AiidaException):
    """
    Raised if attempting to access a locked profile
    """


class LockingProfileError(AiidaException):
    """
    Raised if the profile can`t be locked
    """



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`Data` sub class to represent a dictionary."""
import copy

from aiida.common import exceptions

from .base import to_aiida_type
from .data import Data

__all__ = ('Dict',)


class Dict(Data):
    """`Data` sub class to represent a dictionary.

    The dictionary contents of a `Dict` node are stored in the database as attributes. The dictionary
    can be initialized through the `dict` argument in the constructor. After construction, values can
    be retrieved and updated through the item getters and setters, respectively:

        node['key'] = 'value'

    Alternatively, the `dict` property returns an instance of the `AttributeManager` that can be used
    to get and set values through attribute notation:

        node.dict.key = 'value'

    Note that trying to set dictionary values directly on the node, e.g. `node.key = value`, will not
    work as intended. It will merely set the `key` attribute on the node instance, but will not be
    stored in the database. As soon as the node goes out of scope, the value will be lost.

    It is also relevant to note here the difference in something being an "attribute of a node" (in
    the sense that it is stored in the "attribute" column of the database when the node is stored)
    and something being an "attribute of a python object" (in the sense of being able to modify and
    access it as if it was a property of the variable, e.g. `node.key = value`). This is true of all
    types of nodes, but it becomes more relevant for `Dict` nodes where one is constantly manipulating
    these attributes.

    Finally, all dictionary mutations will be forbidden once the node is stored.
    """

    def __init__(self, value=None, **kwargs):
        """Initialise a ``Dict`` node instance.

        Usual rules for attribute names apply, in particular, keys cannot start with an underscore, or a ``ValueError``
        will be raised.

        Initial attributes can be changed, deleted or added as long as the node is not stored.

        :param value: dictionary to initialise the ``Dict`` node from
        """
        dictionary = value or kwargs.pop('dict', None)
        super().__init__(**kwargs)
        if dictionary:
            self.set_dict(dictionary)

    def __getitem__(self, key):
        try:
            return self.base.attributes.get(key)
        except AttributeError as exc:
            raise KeyError from exc

    def __setitem__(self, key, value):
        self.base.attributes.set(key, value)

    def __eq__(self, other):
        if isinstance(other, Dict):
            return self.get_dict() == other.get_dict()
        return self.get_dict() == other

    def __contains__(self, key: str) -> bool:
        """Return whether the node contains a key."""
        return key in self.base.attributes

    def set_dict(self, dictionary):
        """Replace the current dictionary with another one.

        :param dictionary: dictionary to set
        """
        dictionary_backup = copy.deepcopy(self.get_dict())

        try:
            # Clear existing attributes and set the new dictionary
            self.base.attributes.clear()
            self.update_dict(dictionary)
        except exceptions.ModificationNotAllowed:  # pylint: disable=try-except-raise
            # I reraise here to avoid to go in the generic 'except' below that would raise the same exception again
            raise
        except Exception:
            # Try to restore the old data
            self.base.attributes.clear()
            self.update_dict(dictionary_backup)
            raise

    def update_dict(self, dictionary):
        """Update the current dictionary with the keys provided in the dictionary.

        .. note:: works exactly as `dict.update()` where new keys are simply added and existing keys are overwritten.

        :param dictionary: a dictionary with the keys to substitute
        """
        for key, value in dictionary.items():
            self.base.attributes.set(key, value)

    def get_dict(self):
        """Return a dictionary with the parameters currently set.

        :return: dictionary
        """
        return dict(self.base.attributes.all)

    def keys(self):
        """Iterator of valid keys stored in the Dict object.

        :return: iterator over the keys of the current dictionary
        """
        for key in self.base.attributes.keys():
            yield key

    def items(self):
        """Iterator of all items stored in the Dict node."""
        for key, value in self.base.attributes.items():
            yield key, value

    @property
    def dict(self):
        """Return an instance of `AttributeManager` that transforms the dictionary into an attribute dict.

        .. note:: this will allow one to do `node.dict.key` as well as `node.dict[key]`.

        :return: an instance of the `AttributeResultManager`.
        """
        from aiida.orm.utils.managers import AttributeManager
        return AttributeManager(self)


@to_aiida_type.register(dict)
def _(value):
    return Dict(value)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""The main `verdi` click group."""
import difflib

import click
from click import shell_completion

from aiida import __version__
from aiida.cmdline.params import options, types

GIU = (
    'ABzY8%U8Kw0{@klyK?I~3`Ki?#qHQ&IIM|J;6yB`9_+{&w)p(JK}vokj-11jhve8xcx?dZ>+9nwrEF!x*S>9A+EWYrR?6GA-u?jFa+et65GF@1+D{%'
    '8{C~xjt%>uVM4RTSS?j2M)XH%T#>M{K$lE2XGD`<aYaOFUit|Wh*Q?-n)Bs}n0}hc2!oAkoBQyEOaSkqfd=oq;sTs(x@(H&qOKnXDFA52~gq4CvNbk'
    '%p9&pE+KH!2lm^MThvE$xC2x*>RS0T67213wbAs!SZmn+;(-m!>f(T@e%@oxd`yRBp9nu+9N`4xv8AS@O$CaQ;7FXzM=ug^$?3ta2551EDL`wK4|Cm'
    '%RnJdS#0UF<by)B<XQ+8vZ}j$b$@4Q1#kj{G|g6{1#bLd+w)SZ16_wTnfo<QXMd-s4ImikcvSj#7~2Yq-1l@K)y|aS%KqzrV$m)VgR$xs8bGr7%$lJ'
    '_ObpObgS_mhZ7M97P}ATen9Zil^7;P;<I08Mv_;!$Eb2P)F|&t+wT?>wVwe<Fez}g$`Zi>DkcfdNjtUv1N^iSQui#TL(q!FmIeKb!yW4|L`@!@-4x6'
    'B6I^ptRdH+4o0ODM;1_f^}4@LMe@#_YHz0wQdq@d)@n)uYNtAb2OLo&fpBkct5{~3kbRag^_5QG%qrTksHMXAYAQoz1#2wtHCy0}h?CJtzv&@Q?^9r'
    'd&02;isB7NJMMr7F@>$!ELj(sbwzIR4)rnch=oVZrG;8)%R6}FUk*fv2O&!#ZA)$HloK9!es&4Eb+h=OIyWFha(8PPy9u?NqfkuPYg;GO1RVzBLX)7'
    'ORMM>1hEM`-96mGjJ+A!e-_}4X{M|4CkKE~uF4j+LW#6IsFa*_da_mLqzr)E<`%ikthkMO2<vdLNlWMLBrceLb&%p<SlbT6NAh+Tz~72^U2(&}V&4H'
    'd8r#{;M;(*9swiCZ2RIx2&yLd0wV^AQs5$V63{q89vFwvV_?Vk!HuDTPr83yG~Wm}YG_*0gIL7B~{>>65cNMtpDE*VejqZV^MyewPJJAS*VM6jY;QY'
    '#g7gOKgPbFg{@;YDL6Gbxxr|2T&BQunB?PBetq?X<bp0b)qASa|!TspwLp%9CE7Y#XI@}Wa3E6#mZC62W0F#k>>jW1hFF7&>EaYkKYqIa_ld(Z@AJT'
    '+lJ(Pd;+?<&&M>A0agti19^z3n4Z6_WG}c~_+XHyJI_iau7+V$#YA$pJ~H)yHEVy1D?5^Sw`tb@{nnNNo=eSMZLf0>m^A@7f{y$nb_HJWgLRtZ?<Vq'
    'Glx*4Tkeba)%-_D&vG1uUL0(DEK)&;4yGQ@C_0d4x%#)?G?XfN*{PN40X3QeC3@;*9uMwABesu!8?(0#>x2?*>SwM?JoQ>p|-1ZRU0#+{^UhK22+~o'
    'R9k7rh<eeM8~?e5)U+OloQYk-ebZsXE8KFmR5;uE)C;wlw}@dIbx-{${l+HbC|PrK*6{Q(q6jMpni4Be``)PJJRU>(GH9y|jm){jY9_xAI4N_EfU#4'
    'taTUXFY4a4l$v=N-+f+w&wuH;Z(6p6#=n8XwlZ;*L&-rcL~T_vEm@#-Xi8&g06!MO+R(<NRBkE$(0Y_~^2C_k<o~_a3*HAHukZKXCs8^y%UZZVvze'
)


def _start_of_option(value: str) -> bool:
    """Check if the value looks like the start of an option.

    This is an adaptation of :py:func:`click.shell_completion._start_of_option` that simply add ``.``, ``~``, ``$`` as
    the characters that are interpreted as the start of a filepath, and so not the start of an option. This will ensure
    that filepaths starting with these characters are autocompleted once again.

    Here ``.`` indicates a relative path, ``~`` indicates the home directory, and ``$`` allows to expand environment
    variables such as ``$HOME`` and ``$PWD``.
    """
    if not value:
        return False

    # Allow characters that typically designate the start of a path.
    return not value[0].isalnum() and value[0] not in ['/', '.', '~', '$']


shell_completion._start_of_option = _start_of_option  # pylint: disable=protected-access


class VerdiCommandGroup(click.Group):
    """Custom class for ``verdi`` top-level command group."""

    @staticmethod
    def add_verbosity_option(cmd):
        """Apply the ``verbosity`` option to the command, which is common to all ``verdi`` commands."""
        # Only apply the option if it hasn't been already added in a previous call.
        if cmd is not None and 'verbosity' not in [param.name for param in cmd.params]:
            cmd = options.VERBOSITY()(cmd)

        return cmd

    def fail_with_suggestions(self, ctx, cmd_name):
        """Fail the command while trying to suggest commands to resemble the requested ``cmd_name``."""
        # We might get better results with the Levenshtein distance or more advanced methods implemented in FuzzyWuzzy
        # or similar libs, but this is an easy win for now.
        matches = difflib.get_close_matches(cmd_name, self.list_commands(ctx), cutoff=0.5)

        if not matches:
            # Single letters are sometimes not matched so also try with a simple startswith
            matches = [c for c in sorted(self.list_commands(ctx)) if c.startswith(cmd_name)][:3]

        if matches:
            formatted = '\n'.join(f'\t{m}' for m in sorted(matches))
            ctx.fail(f'`{cmd_name}` is not a {self.name} command.\n\nThe most similar commands are:\n{formatted}')
        else:
            ctx.fail(f'`{cmd_name}` is not a {self.name} command.\n\nNo similar commands found.')

    def get_command(self, ctx, cmd_name):
        """Return the command that corresponds to the requested ``cmd_name``.

        This method is overridden from the base class in order to two functionalities:

            * If the command is found, automatically add the verbosity option.
            * If the command is not found, attempt to provide a list of suggestions with existing commands that resemble
              the requested command name.

        Note that if the command is not found and ``resilient_parsing`` is set to True on the context, then the latter
        feature is disabled because most likely we are operating in tab-completion mode.
        """
        if int(cmd_name.lower().encode('utf-8').hex(), 16) == 0x6769757365707065:
            import base64
            import gzip
            click.echo(gzip.decompress(base64.b85decode(GIU.encode('utf-8'))).decode('utf-8'))
            return None

        cmd = super().get_command(ctx, cmd_name)

        if cmd is not None:
            return self.add_verbosity_option(cmd)

        # If this command is called during tab-completion, we do not want to print an error message if the command can't
        # be found, but instead we want to simply return here. However, in a normal command execution, we do want to
        # execute the rest of this method to try and match commands that are similar in order to provide the user with
        # some hints. The problem is that there is no one canonical way to determine whether the invocation is due to a
        # normal command execution or a tab-complete operation. The `resilient_parsing` attribute of the `Context` is
        # designed to allow things like tab-completion, however, it is not the only purpose. For now this is our best
        # bet though to detect a tab-complete event. When `resilient_parsing` is switched on, we assume a tab-complete
        # and do nothing in case the command name does not match an actual command.
        if ctx.resilient_parsing:
            return

        self.fail_with_suggestions(ctx, cmd_name)

    def group(self, *args, **kwargs):
        """Ensure that sub command groups use the same class but do not override an explicitly set value."""
        kwargs.setdefault('cls', self.__class__)
        return super().group(*args, **kwargs)


# Pass the version explicitly to ``version_option`` otherwise editable installs can show the wrong version number
@click.command(cls=VerdiCommandGroup, context_settings={'help_option_names': ['--help']})
@options.PROFILE(type=types.ProfileParamType(load_profile=True), expose_value=False)
@options.VERBOSITY()
@click.version_option(__version__, package_name='aiida_core', message='AiiDA version %(version)s')
def verdi():
    """The command line interface of AiiDA."""



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=cyclic-import
"""
.. py:module::overridable
    :synopsis: Convenience class which can be used to defined a set of commonly used options that
        can be easily reused and which improves consistency across the command line interface
"""

import click

__all__ = ('OverridableOption',)


class OverridableOption:
    """
    Wrapper around click option that increases reusability

    Click options are reusable already but sometimes it can improve the user interface to for example customize a
    help message for an option on a per-command basis. Sometimes the option should be prompted for if it is not given
    On some commands an option might take any folder path, while on another the path only has to exist.

    Overridable options store the arguments to click.option and only instantiate the click.Option on call,
    kwargs given to ``__call__`` override the stored ones.

    Example::

        FOLDER = OverridableOption('--folder', type=click.Path(file_okay=False), help='A folder')

        @click.command()
        @FOLDER(help='A folder, will be created if it does not exist')
        def ls_or_create(folder):
            click.echo(os.listdir(folder))

        @click.command()
        @FOLDER(help='An existing folder', type=click.Path(exists=True, file_okay=False, readable=True)
        def ls(folder)
            click.echo(os.listdir(folder))
    """

    def __init__(self, *args, **kwargs):
        """
        Store the default args and kwargs.

        :param args: default arguments to be used for the click option
        :param kwargs: default keyword arguments to be used that can be overridden in the call
        """
        self.args = args
        self.kwargs = kwargs

    def __call__(self, **kwargs):
        """
        Override the stored kwargs, (ignoring args as we do not allow option name changes) and return the option.

        :param kwargs: keyword arguments that will override those set in the construction
        :return: click option constructed with args and kwargs defined during construction and call of this instance
        """
        kw_copy = self.kwargs.copy()
        kw_copy.update(kwargs)
        return click.option(*self.args, **kw_copy)

    def clone(self, **kwargs):
        """
        Create a new instance of the OverridableOption by cloning it and updating the stored kwargs with those passed.

        This can be useful when an already predefined OverridableOption needs to be further specified and reused
        by a set of sub commands. Example::

            LABEL = OverridableOption('-l', '--label', required=False, help='The label of the node'
            LABEL_COMPUTER = LABEL.clone(required=True, help='The label of the computer')

        If multiple computer related sub commands need the LABEL option, but the default help string and required
        attribute need to be different, the `clone` method allows to override these and create a new OverridableOption
        instance that can then be used as a decorator.

        :param kwargs: keyword arguments to update
        :return: OverridableOption instance with stored keyword arguments updated
        """
        import copy
        clone = copy.deepcopy(self)
        clone.kwargs.update(kwargs)
        return clone



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi config` command."""
import json
from pathlib import Path
import textwrap

import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import arguments
from aiida.cmdline.utils import echo
from aiida.manage.configuration import MIGRATIONS, downgrade_config, get_config_path
from aiida.manage.configuration.settings import DEFAULT_CONFIG_INDENT_SIZE


@verdi.group('config')
def verdi_config():
    """Manage the AiiDA configuration."""


@verdi_config.command('list')
@click.argument('prefix', metavar='PREFIX', required=False, default='')
@click.option('-d', '--description', is_flag=True, help='Include description of options')
@click.pass_context
def verdi_config_list(ctx, prefix, description: bool):
    """List AiiDA options for the current profile.

    Optionally filtered by a prefix.
    """
    from tabulate import tabulate

    from aiida.manage.configuration import Config, Profile

    config: Config = ctx.obj.config
    profile: Profile = ctx.obj.profile

    if not profile:
        echo.echo_warning('no profiles configured: run `verdi setup` to create one')

    option_values = config.get_options(profile.name if profile else None)

    def _join(val):
        """split arrays into multiple lines."""
        if isinstance(val, list):
            return '\n'.join(str(v) for v in val)
        return val

    if description:
        table = [[name, source, _join(value), '\n'.join(textwrap.wrap(c.description))]
                 for name, (c, source, value) in option_values.items()
                 if name.startswith(prefix)]
        headers = ['name', 'source', 'value', 'description']
    else:
        table = [[name, source, _join(value)]
                 for name, (c, source, value) in option_values.items()
                 if name.startswith(prefix)]
        headers = ['name', 'source', 'value']

    # sort by name
    table = sorted(table, key=lambda x: x[0])
    echo.echo(tabulate(table, headers=headers))


@verdi_config.command('show')
@arguments.CONFIG_OPTION(metavar='OPTION_NAME')
@click.pass_context
def verdi_config_show(ctx, option):
    """Show details of an AiiDA option for the current profile."""
    from aiida.manage.configuration import Config, Profile
    from aiida.manage.configuration.options import NO_DEFAULT

    config: Config = ctx.obj.config
    profile: Profile = ctx.obj.profile

    dct = {
        'schema': option.schema,
        'values': {
            'default': '<NOTSET>' if option.default is NO_DEFAULT else option.default,
            'global': config.options.get(option.name, '<NOTSET>'),
        }
    }

    if not profile:
        echo.echo_warning('no profiles configured: run `verdi setup` to create one')
    else:
        dct['values']['profile'] = profile.options.get(option.name, '<NOTSET>')

    echo.echo_dictionary(dct, fmt='yaml', sort_keys=False)


@verdi_config.command('get')
@arguments.CONFIG_OPTION(metavar='OPTION_NAME')
def verdi_config_get(option):
    """Get the value of an AiiDA option for the current profile."""
    from aiida import get_config_option

    value = get_config_option(option.name)
    echo.echo(str(value))


@verdi_config.command('set')
@arguments.CONFIG_OPTION(metavar='OPTION_NAME')
@click.argument('value', metavar='OPTION_VALUE')
@click.option('-g', '--global', 'globally', is_flag=True, help='Apply the option configuration wide.')
@click.option('-a', '--append', is_flag=True, help='Append the value to an existing array.')
@click.option('-r', '--remove', is_flag=True, help='Remove the value from an existing array.')
@click.pass_context
def verdi_config_set(ctx, option, value, globally, append, remove):
    """Set an AiiDA option.

    List values are split by whitespace, e.g. "a b" becomes ["a", "b"].
    """
    from aiida.manage.configuration import Config, ConfigValidationError, Profile

    if append and remove:
        echo.echo_critical('Cannot flag both append and remove')

    config: Config = ctx.obj.config
    profile: Profile = ctx.obj.profile

    if option.global_only:
        globally = True

    # Define the string that determines the scope: for specific profile or globally
    scope = profile.name if (not globally and profile) else None
    scope_text = f"for '{profile.name}' profile" if (not globally and profile) else 'globally'

    if append or remove:
        try:
            current = config.get_option(option.name, scope=scope)
        except ConfigValidationError as error:
            echo.echo_critical(str(error))
        if not isinstance(current, list):
            echo.echo_critical(f'cannot append/remove to value: {current}')
        if append:
            value = current + [value]
        else:
            value = [item for item in current if item != value]

    # Set the specified option
    try:
        value = config.set_option(option.name, value, scope=scope)
    except ConfigValidationError as error:
        echo.echo_critical(str(error))

    config.store()
    echo.echo_success(f"'{option.name}' set to {value} {scope_text}")


@verdi_config.command('unset')
@arguments.CONFIG_OPTION(metavar='OPTION_NAME')
@click.option('-g', '--global', 'globally', is_flag=True, help='Unset the option configuration wide.')
@click.pass_context
def verdi_config_unset(ctx, option, globally):
    """Unset an AiiDA option."""
    from aiida.manage.configuration import Config, Profile

    config: Config = ctx.obj.config
    profile: Profile = ctx.obj.profile

    if option.global_only:
        globally = True

    # Define the string that determines the scope: for specific profile or globally
    scope = profile.name if (not globally and profile) else None
    scope_text = f"for '{profile.name}' profile" if (not globally and profile) else 'globally'

    # Unset the specified option
    config.unset_option(option.name, scope=scope)
    config.store()
    echo.echo_success(f"'{option.name}' unset {scope_text}")


@verdi_config.command('caching')
@click.option('-d', '--disabled', is_flag=True, help='List disabled types instead.')
def verdi_config_caching(disabled):
    """List caching-enabled process types for the current profile."""
    from aiida.manage.caching import get_use_cache
    from aiida.plugins.entry_point import ENTRY_POINT_STRING_SEPARATOR, get_entry_point_names

    for group in ['aiida.calculations', 'aiida.workflows']:
        for entry_point in get_entry_point_names(group):
            identifier = ENTRY_POINT_STRING_SEPARATOR.join([group, entry_point])
            if get_use_cache(identifier=identifier):
                if not disabled:
                    echo.echo(identifier)
            elif disabled:
                echo.echo(identifier)


@verdi_config.command('downgrade')
@click.argument('version', type=click.Choice({str(m.down_revision) for m in MIGRATIONS}))
def verdi_config_downgrade(version):
    """Print a configuration, downgraded to a specific version."""
    path = Path(get_config_path())
    echo.echo_report(f'Downgrading configuration to v{version}: {path}')
    config = json.loads(path.read_text(encoding='utf8'))
    downgrade_config(config, int(version))
    path.write_text(json.dumps(config, indent=DEFAULT_CONFIG_INDENT_SIZE), encoding='utf8')
    echo.echo_success('Downgraded')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
AiiDA is a flexible and scalable informatics' infrastructure to manage,
preserve, and disseminate the simulations, data, and workflows of
modern-day computational science.

Able to store the full provenance of each object, and based on a
tailored database built for efficient data mining of heterogeneous results,
AiiDA gives the user the ability to interact seamlessly with any number of
remote HPC resources and codes, thanks to its flexible plugin interface
and workflow engine for the automation of complex sequences of simulations.

More information at http://www.aiida.net
"""
from aiida.common.log import configure_logging
from aiida.manage.configuration import get_config_option, get_profile, load_profile, profile_context

__copyright__ = (
    'Copyright (c), This file is part of the AiiDA platform. '
    'For further information please visit http://www.aiida.net/. All rights reserved.'
)
__license__ = 'MIT license, see LICENSE.txt file.'
__version__ = '2.0.0'
__authors__ = 'The AiiDA team.'
__paper__ = (
    'S. P. Huber et al., "AiiDA 1.0, a scalable computational infrastructure for automated reproducible workflows and '
    'data provenance", Scientific Data 7, 300 (2020); https://doi.org/10.1038/s41597-020-00638-4'
)
__paper_short__ = 'S. P. Huber et al., Scientific Data 7, 300 (2020).'


def get_strict_version():
    """
    Return a distutils StrictVersion instance with the current distribution version

    :returns: StrictVersion instance with the current version
    :rtype: :class:`!distutils.version.StrictVersion`
    """
    from distutils.version import StrictVersion
    return StrictVersion(__version__)


def get_version() -> str:
    """
    Return the current AiiDA distribution version

    :returns: the current version
    """
    return __version__


def _get_raw_file_header() -> str:
    """
    Get the default header for source AiiDA source code files.
    Note: is not preceded by comment character.

    :return: default AiiDA source file header
    """
    return f"""This file has been created with AiiDA v. {__version__}
If you use AiiDA for publication purposes, please cite:
{__paper__}
"""


def get_file_header(comment_char: str = '# ') -> str:
    """
    Get the default header for source AiiDA source code files.

    .. note::

        Prepend by comment character.

    :param comment_char: string put in front of each line

    :return: default AiiDA source file header
    """
    lines = _get_raw_file_header().splitlines()
    return '\n'.join(f'{comment_char}{line}' for line in lines)


def load_ipython_extension(ipython):
    """Load the AiiDA IPython extension, using ``%load_ext aiida``."""
    from .tools.ipython.ipython_magics import AiiDALoaderMagics
    ipython.register_magics(AiiDALoaderMagics)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Parity with Django backend (rev: 0048),
part 2: Alter columns to be non-nullable and change type of some columns.

Revision ID: 1de112340b17
Revises: 1de112340b16
Create Date: 2021-08-25 04:28:52.102767

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSONB, UUID

# revision identifiers, used by Alembic.
revision = '1de112340b17'
down_revision = '1de112340b16'
branch_labels = None
depends_on = None


def upgrade():
    """Upgrade database schema."""
    op.alter_column('db_dbauthinfo', 'aiidauser_id', existing_type=sa.INTEGER(), nullable=False)
    op.alter_column('db_dbauthinfo', 'dbcomputer_id', existing_type=sa.INTEGER(), nullable=False)
    op.alter_column('db_dbauthinfo', 'metadata', existing_type=JSONB, nullable=False)
    op.alter_column('db_dbauthinfo', 'auth_params', existing_type=JSONB, nullable=False)
    op.alter_column('db_dbauthinfo', 'enabled', existing_type=sa.BOOLEAN(), nullable=False)

    op.alter_column('db_dbcomment', 'dbnode_id', existing_type=sa.INTEGER(), nullable=False)
    op.alter_column('db_dbcomment', 'user_id', existing_type=sa.INTEGER(), nullable=False)
    op.alter_column('db_dbcomment', 'content', existing_type=sa.TEXT(), nullable=False)
    op.alter_column('db_dbcomment', 'ctime', existing_type=sa.DateTime(timezone=True), nullable=False)
    op.alter_column('db_dbcomment', 'mtime', existing_type=sa.DateTime(timezone=True), nullable=False)
    op.alter_column('db_dbcomment', 'uuid', existing_type=UUID(as_uuid=True), nullable=False)

    op.alter_column('db_dbcomputer', 'description', existing_type=sa.TEXT(), nullable=False)
    op.alter_column('db_dbcomputer', 'hostname', existing_type=sa.String(255), nullable=False)
    op.alter_column('db_dbcomputer', 'metadata', existing_type=JSONB, nullable=False)
    op.alter_column('db_dbcomputer', 'scheduler_type', existing_type=sa.String(255), nullable=False)
    op.alter_column('db_dbcomputer', 'transport_type', existing_type=sa.String(255), nullable=False)
    op.alter_column('db_dbcomputer', 'uuid', existing_type=UUID(as_uuid=True), nullable=False)

    op.alter_column('db_dbgroup', 'user_id', existing_type=sa.INTEGER(), nullable=False)
    op.alter_column('db_dbgroup', 'description', existing_type=sa.TEXT(), nullable=False)
    op.alter_column('db_dbgroup', 'label', existing_type=sa.String(255), nullable=False)
    op.alter_column('db_dbgroup', 'time', existing_type=sa.DateTime(timezone=True), nullable=False)
    op.alter_column('db_dbgroup', 'type_string', existing_type=sa.String(255), nullable=False)
    op.alter_column('db_dbgroup', 'uuid', existing_type=UUID(as_uuid=True), nullable=False)

    op.alter_column('db_dbgroup_dbnodes', 'dbnode_id', existing_type=sa.INTEGER(), nullable=False)
    op.alter_column('db_dbgroup_dbnodes', 'dbgroup_id', existing_type=sa.INTEGER(), nullable=False)

    op.alter_column('db_dblink', 'type', existing_type=sa.String(255), nullable=False)
    op.alter_column('db_dblink', 'input_id', existing_type=sa.INTEGER(), nullable=False)
    op.alter_column('db_dblink', 'output_id', existing_type=sa.INTEGER(), nullable=False)

    op.alter_column('db_dblog', 'levelname', existing_type=sa.String(255), type_=sa.String(50), nullable=False)
    op.alter_column('db_dblog', 'loggername', existing_type=sa.String(255), nullable=False)
    op.alter_column('db_dblog', 'message', existing_type=sa.TEXT(), nullable=False)
    op.alter_column('db_dblog', 'time', existing_type=sa.DateTime(timezone=True), nullable=False)
    op.alter_column('db_dblog', 'uuid', existing_type=UUID(as_uuid=True), nullable=False)
    op.alter_column('db_dblog', 'metadata', existing_type=JSONB, nullable=False)

    op.alter_column('db_dbnode', 'ctime', existing_type=sa.DateTime(timezone=True), nullable=False)
    op.alter_column('db_dbnode', 'description', existing_type=sa.TEXT(), nullable=False)
    op.alter_column('db_dbnode', 'label', existing_type=sa.String(255), nullable=False)
    op.alter_column('db_dbnode', 'mtime', existing_type=sa.DateTime(timezone=True), nullable=False)
    op.alter_column('db_dbnode', 'node_type', existing_type=sa.String(255), nullable=False)
    op.alter_column('db_dbnode', 'uuid', existing_type=UUID(as_uuid=True), nullable=False)

    op.alter_column('db_dbsetting', 'time', existing_type=sa.DateTime(timezone=True), nullable=False)
    op.alter_column('db_dbsetting', 'key', existing_type=sa.String(255), type_=sa.String(1024), nullable=False)
    op.alter_column('db_dbsetting', 'description', existing_type=sa.String(255), type_=sa.Text(), nullable=False)

    op.alter_column('db_dbuser', 'email', existing_type=sa.String(254), nullable=False)
    op.alter_column('db_dbuser', 'first_name', existing_type=sa.String(254), nullable=False)
    op.alter_column('db_dbuser', 'last_name', existing_type=sa.String(254), nullable=False)
    op.alter_column('db_dbuser', 'institution', existing_type=sa.String(254), nullable=False)


def downgrade():
    """Downgrade database schema."""
    op.alter_column('db_dbuser', 'institution', existing_type=sa.String(254), nullable=True)
    op.alter_column('db_dbuser', 'last_name', existing_type=sa.String(254), nullable=True)
    op.alter_column('db_dbuser', 'first_name', existing_type=sa.String(254), nullable=True)
    op.alter_column('db_dbuser', 'email', existing_type=sa.String(254), nullable=True)

    op.alter_column('db_dbsetting', 'time', existing_type=sa.DateTime(timezone=True), nullable=True)
    op.alter_column('db_dbsetting', 'key', existing_type=sa.String(1024), type_=sa.String(255), nullable=False)
    op.alter_column('db_dbsetting', 'description', existing_type=sa.Text(), type_=sa.String(255), nullable=False)

    op.alter_column('db_dbnode', 'ctime', existing_type=sa.DateTime(timezone=True), nullable=True)
    op.alter_column('db_dbnode', 'description', existing_type=sa.TEXT(), nullable=True)
    op.alter_column('db_dbnode', 'label', existing_type=sa.String(255), nullable=True)
    op.alter_column('db_dbnode', 'mtime', existing_type=sa.DateTime(timezone=True), nullable=True)
    op.alter_column('db_dbnode', 'node_type', existing_type=sa.String(255), nullable=True)
    op.alter_column('db_dbnode', 'uuid', existing_type=UUID(as_uuid=True), nullable=True)

    op.alter_column('db_dblog', 'metadata', existing_type=JSONB, nullable=True)
    op.alter_column('db_dblog', 'message', existing_type=sa.TEXT(), nullable=True)
    op.alter_column('db_dblog', 'levelname', existing_type=sa.String(50), type_=sa.String(255), nullable=True)
    op.alter_column('db_dblog', 'loggername', existing_type=sa.String(255), nullable=True)
    op.alter_column('db_dblog', 'time', existing_type=sa.DateTime(timezone=True), nullable=True)
    op.alter_column('db_dblog', 'uuid', existing_type=UUID(as_uuid=True), nullable=True)

    op.alter_column('db_dblink', 'output_id', existing_type=sa.INTEGER(), nullable=True)
    op.alter_column('db_dblink', 'input_id', existing_type=sa.INTEGER(), nullable=True)
    op.alter_column('db_dblink', 'type', existing_type=sa.String(255), nullable=True)

    op.alter_column('db_dbgroup_dbnodes', 'dbgroup_id', existing_type=sa.INTEGER(), nullable=True)
    op.alter_column('db_dbgroup_dbnodes', 'dbnode_id', existing_type=sa.INTEGER(), nullable=True)

    op.alter_column('db_dbgroup', 'user_id', existing_type=sa.INTEGER(), nullable=True)
    op.alter_column('db_dbgroup', 'description', existing_type=sa.TEXT(), nullable=True)
    op.alter_column('db_dbgroup', 'time', existing_type=sa.DateTime(timezone=True), nullable=True)
    op.alter_column('db_dbgroup', 'type_string', existing_type=sa.String(255), nullable=True)
    op.alter_column('db_dbgroup', 'label', existing_type=sa.String(255), nullable=True)
    op.alter_column('db_dbgroup', 'uuid', existing_type=UUID(as_uuid=True), nullable=True)

    op.alter_column('db_dbcomputer', 'metadata', existing_type=JSONB, nullable=True)
    op.alter_column('db_dbcomputer', 'transport_type', existing_type=sa.String(255), nullable=True)
    op.alter_column('db_dbcomputer', 'scheduler_type', existing_type=sa.String(255), nullable=True)
    op.alter_column('db_dbcomputer', 'description', existing_type=sa.TEXT(), nullable=True)
    op.alter_column('db_dbcomputer', 'hostname', existing_type=sa.String(255), nullable=True)
    op.alter_column('db_dbcomputer', 'uuid', existing_type=UUID(as_uuid=True), nullable=True)

    op.alter_column('db_dbcomment', 'user_id', existing_type=sa.INTEGER(), nullable=True)
    op.alter_column('db_dbcomment', 'dbnode_id', existing_type=sa.INTEGER(), nullable=True)
    op.alter_column('db_dbcomment', 'content', existing_type=sa.TEXT(), nullable=True)
    op.alter_column('db_dbcomment', 'ctime', existing_type=sa.DateTime(timezone=True), nullable=True)
    op.alter_column('db_dbcomment', 'mtime', existing_type=sa.DateTime(timezone=True), nullable=True)
    op.alter_column('db_dbcomment', 'uuid', existing_type=UUID(as_uuid=True), nullable=True)

    op.alter_column('db_dbauthinfo', 'dbcomputer_id', existing_type=sa.INTEGER(), nullable=True)
    op.alter_column('db_dbauthinfo', 'aiidauser_id', existing_type=sa.INTEGER(), nullable=True)
    op.alter_column('db_dbauthinfo', 'enabled', existing_type=sa.BOOLEAN(), nullable=True)
    op.alter_column('db_dbauthinfo', 'auth_params', existing_type=JSONB, nullable=True)
    op.alter_column('db_dbauthinfo', 'metadata', existing_type=JSONB, nullable=True)
